/* Sample solution for a very old assignment on   */
/* call routing in circuit-switched networks.     */
/* Written by Carey Williamson, February 26, 2001 */
/*                                                */
/* Usage: cc -o route route.c                     */
/*        route                                   */
/* Requires:                                      */
/*        topology.dat   - network topology file  */
/*        workload.dat   - workload file          */

#include <stdio.h>
#include <stdlib.h>

/* Network topology and simulation info */
#define MAX_ROW 30
#define MAX_COL 30
#define MAX_EVENTS 50000
#define INFINITY 999999
#define BUSY     888888
#define DELTA   0.001
#define BIGGIE   100

/* Call event types */
#define CALL_ARRIVAL 1
#define CALL_DEPARTURE 2

/* Routing algorithms */
#define SHORTEST_HOP_PATH_FIRST 1
/* #define SHORTEST_HOP_PATH_ONLY 1 */
/* #define SHORTEST_DELAY_PATH_FIRST 1 */
/* #define LEAST_LOADED_PATH 1 */
/* #define MAXIMUM_FREE_CIRCUIT 1 */
#define NEW 1

/* Debugging flag */
/* #define DEBUG 1 */   /* Input files and paths */
/* #define DEBUG2 1 */  /* Details of Dijkstra's algorithm */
/* #define DEBUG3 1 */  /* LLP link costs */

/* Detailed report of Zoom load on link between U and R */
//#define LINK 1

/* Global data structures for network topology and state information */
int activecalls;
int propdelay[MAX_ROW][MAX_COL];
int capacity[MAX_ROW][MAX_COL];
int available[MAX_ROW][MAX_COL];
#ifdef LEAST_LOADED_PATH
float cost[MAX_ROW][MAX_COL];
#else
int cost[MAX_ROW][MAX_COL];
#endif
int mincost[MAX_ROW][MAX_COL];
char route[MAX_ROW];
char routes[MAX_EVENTS][MAX_ROW];
int hops[MAX_EVENTS];

int numnodes, numlinks, numevents, numhops;
int calls, success, blocked;
int pathdelay, totdelay, tothops;
float now;

FILE *fp1, *fp2;

struct Event
  {
    float event_time;
    int event_type;
    int callid;
    char source;
    char destination;
    float duration;
  } EventList[MAX_EVENTS];

int current_event;

extern void qsort();

int mycompare(num1, num2)
     float *num1;
     float *num2;
{
  if( *num1 < *num2 )
    return( -1);
  else if( *num1 > *num2 )
    return( 1);
  else return(0);
}

#ifdef LEAST_LOADED_PATH
float DoDijkstra(from, to)
#else
int DoDijkstra(from, to)
#endif
    char from, to;
  {
    char donewith[MAX_ROW];
    int numdonewith, node, node2;
    char bywayof[MAX_ROW];
    int i, j, closestnode, done;
#ifdef LEAST_LOADED_PATH
    float closestdistance;
    float howfar[MAX_ROW];
    float okay;
#else
    int closestdistance;
    int howfar[MAX_ROW];
    int okay;
#endif
    float fix;

#ifdef DEBUG
    printf("DoDijkstra: trying to route from %c to %c\n", from, to);
#endif

    numdonewith = 0;
    donewith[numdonewith] = from;
    numdonewith++;
    for( i = 0; i < numnodes; i++ )
      {
	if( i + 'A' == from )
	  continue;
#ifdef NEW
#ifdef LEAST_LOADED_PATH
	if( cost[from-'A'][i] == INFINITY )
	  howfar[i] = INFINITY;
	else if( cost[from-'A'][i] >= 1.0 )
#else
	if( cost[from-'A'][i] == BUSY )
#endif
	  {
#ifdef DEBUG2
	    printf("  Link from source %c to %c is completely full right now\n",
		   from, i+'A');
#endif
	    howfar[i] = INFINITY;
	  }
	else 
#endif /* NEW */
	  if( cost[from-'A'][i] < INFINITY )
	  {
	    howfar[i] = cost[from-'A'][i];
	    bywayof[i] = from;
#ifdef DEBUG2
#ifdef LEAST_LOADED_PATH
	    printf("  Node %c is directly reachable from %c in %4.3f\n",
		   i+'A', from, howfar[i]);
#else
	    printf("  Node %c is directly reachable from %c in %d\n",
		   i+'A', from, howfar[i]);
#endif /* LLP */
#endif /* DEBUG2 */
	  }
	else howfar[i] = INFINITY;
      }

    while( numdonewith < numnodes )
      {
	/* Find the node closest to source (from) that is */
	/* not yet in the done with set */
	closestdistance = INFINITY;
	closestnode = -1;
	for( i = 0; i < numnodes; i++ )
	  {
	    if( i + 'A' == from )
	      continue;
	    /* make sure that it is not yet in the donewith set */
	    done = 0;
	    for( j = 0; j < numdonewith; j++ )
	      {
		if( donewith[j] == i + 'A' )
		    done = 1;
	      }
	    if( done )
	      continue;
	    if( howfar[i] < closestdistance )
	      {
		closestdistance = howfar[i];
		closestnode = i;
	      }
	  }
	if( closestdistance >= BUSY )
	  return( 0 );

	if( closestnode < 0 )
	  {
	    printf("Problem!\n");
	  }
#ifdef DEBUG2
	printf("  Chose %c as the closest node not yet done with\n",
	       closestnode+'A');
#endif
	/* Put this node into the donewith set */
	donewith[numdonewith] = closestnode + 'A';
	numdonewith++;
#define HACK 1
#ifdef HACK
	if( closestnode + 'A' == to )
	  break;
#endif

	/* Update distances, if necessary */
	for( i = 0; i < numnodes; i++ )
	  {
	    /* Find an adjacent node... */
	    if( cost[closestnode][i] == INFINITY )
	      continue;
#ifdef NEW
#ifdef LEAST_LOADED_PATH
	    if( cost[closestnode][i] >= 1.0 )
#else
	    if( cost[closestnode][i] == BUSY )
#endif /* LLP */
	      {
#ifdef DEBUG2
		printf("  Link from %c to %c is completely full right now\n",
		       closestnode+'A', i+'A');
#endif /* DEBUG2 */
		continue;
	      }
#endif /* NEW */

	    /* that is not yet in the donewith set */
	    done = 0;
	    for( j = 0; j < numdonewith; j++ )
	      {
		if( donewith[j] == i + 'A' )
		    done = 1;
	      }
	    if( done )
	      continue;

#ifdef LEAST_LOADED_PATH
	    if( howfar[closestnode] > cost[closestnode][i] )
	      fix = howfar[closestnode];
	    else fix = cost[closestnode][i];
	    if( fix < howfar[i] )
	      {
		howfar[i] = fix;
		bywayof[i] = closestnode + 'A';
#ifdef DEBUG2
		printf("Dijkstra chooses to route from %c to %c via %c\n",
		       from, i+'A', closestnode+'A');
#endif
	      }
#else
#ifdef MAXIMUM_FREE_CIRCUIT
	    if( howfar[closestnode] > cost[closestnode][i] )
	      fix = howfar[closestnode];
	    else fix = cost[closestnode][i];
	    if( fix < howfar[i] )
	      {
		howfar[i] = fix;
		bywayof[i] = closestnode + 'A';
#ifdef DEBUG2
		printf("Dijkstra Chooses to route from %c to %c via %c\n",
		       from, i+'A', closestnode+'A');
#endif
	      }
#else
	    if( howfar[closestnode] + cost[closestnode][i] < howfar[i] )
	      {
		howfar[i] = howfar[closestnode] + cost[closestnode][i];
		bywayof[i] = closestnode + 'A';
#ifdef DEBUG2
		printf("Dijkstra chooses to route from %c to %c via %c\n",
		       from, i+'A', closestnode+'A');
#endif /* DEBUG2 */
	      }
#endif /* MFC */
#endif /* LLP */
	  }
      }
    /* Now record the route to use */
#ifdef DEBUG2
    printf("For source node %c, the routes are as follows: \n", from);
    for( i = 0; i < numnodes; i++ )
      {
	if( i + 'A' == from )
	  continue;
#ifdef LEAST_LOADED_PATH
	printf("  Destination %c is reachable in %4.3f by way of %c\n",
	       i+'A', howfar[i], bywayof[i]);
#else
	printf("  Destination %c is reachable in %d by way of %c\n",
	       i+'A', howfar[i], bywayof[i]);
#endif /* LLP */
      }
#endif /* DEBUG2 */

    numhops = 0;
    pathdelay = 0;
    node = to;
    route[numhops] = node;
    numhops++;
    okay = 0;
    while( node != from )
      {
	node2 = bywayof[node-'A'];
	route[numhops] = node2;
	if( available[node-'A'][node2-'A'] <= 0 )
	  {
	    okay = 0;
	    /* 	    printf("%c %c  ", node, node2); */
	  }
#ifdef LEAST_LOADED_PATH
	else if( cost[node-'A'][node2-'A'] > okay )
	  {
#ifdef DEBUG
	    printf("Updating okay from %4.3f to ", okay);
#endif
	    okay = cost[node-'A'][node2-'A'];
#ifdef DEBUG
	    printf("%4.3f\n", okay);
#endif
	  }
#else
	/* New thing to try out */
#ifdef MAXIMUM_FREE_CIRCUIT
	else if( cost[node-'A'][node2-'A'] > okay )
	  {
	    okay = cost[node-'A'][node2-'A'];
	  }
#else
	else okay += cost[node-'A'][node2-'A'];
#endif /* MFC */
#endif /* LLP */
	pathdelay += propdelay[node-'A'][node2-'A'];
	node = node2;
	numhops++;
      }
#ifdef DEBUG
    printf("Route to use is: ");
    for( i = numhops-1; i >= 0; i-- )
      {
	printf(" %c ", route[i]);
      }
    printf("\n");
#endif
    return( okay );
  }

int RouteCall(from, to, callid)
    char from, to;
    int callid;
  {
    int i;
#ifdef LEAST_LOADED_PATH
    float totalcost;
#else
    int totalcost;
#endif

#ifdef DEBUG
    printf("RouteCall: trying to route call %d from %c to %c\n",
	   callid, from, to);
#endif

    if( from == to )
      {
	printf("RouteCall: identical source and destination!\n");
	return(1);
      }

    totalcost = DoDijkstra(from, to);
#ifdef SHORTEST_HOP_PATH_ONLY
    if( totalcost > mincost[from-'A'][to-'A'] )
      {
#ifdef DEBUG
	printf("Whoops! Can't use a suboptimal path (%d > %d). Sorry!\n",
	       totalcost, mincost[from-'A'][to-'A']);
#endif
	totalcost = 0;
      }
#endif /* SHPO */

/* #define NO_HOGGIES1 1 */
/* Small hack to MFC to prevent excessive number of hops */
#ifdef NO_HOGGIES1
    if( numhops-1 > 10 )
      totalcost = 0;
#endif

/* #define NO_HOGGIES2 1 */
/* Small hack to MFC to prevent excessive duration calls */
/* This one seems to be the most effective.   CLW  */
#ifdef NO_HOGGIES2
    if( EventList[current_event].duration > 300.0 )
      totalcost = 0;
#endif
    
/* #define NO_HOGGIES3 1 */
/* Small hack to MFC to prevent excessive number of hops and duration */
#ifdef NO_HOGGIES3
    if( (numhops-1 > 8) && (EventList[current_event].duration > 200.0) )
      totalcost = 0;
#endif
    
    if( totalcost > 0 )
      {
#ifdef DEBUG
#ifdef LEAST_LOADED_PATH
	printf("Using a path of cost %4.3f\n", totalcost);
#else
	printf("Using a path with hops %d and cost %d\n", numhops-1, totalcost);
#endif
#endif /* DEBUG */
	/* Record the route for future reference */
	for( i = numhops-1; i >= 0; i-- )
	  {
	    routes[callid][i] = route[i];
	    hops[callid] = numhops;
	  }
	/* Record the existence of this call consuming resources */
	for( i = numhops-1; i >= 0; i-- )
	  {
	    char node1, node2;
	    node1 = route[i];
	    if( i > 0 )
	      {
		float temp;
		node2 = route[i-1];
#ifdef LINK
		if( (node1 == 'U') && (node2 == 'R') )
		  printf("%f %d\n",
			 now, capacity[node1-'A'][node2-'A'] - available[node1-'A'][node2-'A']);
#endif
		available[node1-'A'][node2-'A'] -= 1;
		available[node2-'A'][node1-'A'] -= 1;
		if( available[node1-'A'][node2-'A'] == 0 )
		  printf("Link from %c to %c is now FULL!!\n", node1, node2);
#ifdef LINK
		if( (node1 == 'U') && (node2 == 'R') )
		  printf("%f %d\n",
			 now, capacity[node1-'A'][node2-'A'] - available[node1-'A'][node2-'A']);
#endif
#ifdef NEW
#ifdef LEAST_LOADED_PATH
		temp = capacity[node2-'A'][node1-'A'];
		temp -= available[node2-'A'][node1-'A'];
		temp /= capacity[node2-'A'][node1-'A'];
		temp += DELTA;
#ifdef DEBUG3
		printf("Setting link cost from %c to %c to %4.3f\n",
		       node2, node1, temp);
#endif
		cost[node2-'A'][node1-'A'] = temp;
		cost[node1-'A'][node2-'A'] = temp;
#else /* LLP */
		if( available[node2-'A'][node1-'A'] == 0 )
		  {
		    cost[node2-'A'][node1-'A'] = BUSY;
		    cost[node1-'A'][node2-'A'] = BUSY;
		  }
#endif /* LLP */
#ifdef MAXIMUM_FREE_CIRCUIT
	    cost[node2-'A'][node1-'A'] = BIGGIE - available[node2-'A'][node1-'A'];
	    cost[node1-'A'][node2-'A'] = BIGGIE - available[node1-'A'][node2-'A'];
#endif /* MFC */
#endif /* NEW */
	      }
	  }
	return(1);
      }
    else return(0);
  }

int ReleaseCall(from, to, callid)
    char from, to;
    int callid;
  {
    int i;
    numhops = hops[callid];
    for( i = numhops-1; i >= 0; i-- )
      {
	char node1, node2;
	node1 = routes[callid][i];
	if( i > 0 )
	  {
	    float temp;
	    node2 = routes[callid][i-1];
	    available[node1-'A'][node2-'A'] += 1;
	    available[node2-'A'][node1-'A'] += 1;
#ifdef NEW
#ifdef LEAST_LOADED_PATH
	    temp = capacity[node2-'A'][node1-'A'];
	    temp -= available[node2-'A'][node1-'A'];
	    temp /= capacity[node2-'A'][node1-'A'];
	    temp += DELTA;
#ifdef DEBUG3
	    printf("Setting link cost from %c to %c to %4.3f\n",
		   node2, node1, temp);
#endif
	    cost[node2-'A'][node1-'A'] = temp;
	    cost[node1-'A'][node2-'A'] = temp;
#else /* LLP */
#endif /* LLP */
	    if( available[node2-'A'][node1-'A'] == 1 )
	      {
#ifdef SHORTEST_DELAY_PATH_FIRST
		cost[node2-'A'][node1-'A'] = propdelay[node2-'A'][node1-'A'];
		cost[node1-'A'][node2-'A'] = propdelay[node1-'A'][node2-'A'];
#endif
#ifdef SHORTEST_HOP_PATH_FIRST
		cost[node2-'A'][node1-'A'] = 1;
		cost[node1-'A'][node2-'A'] = 1;
#endif
#ifdef SHORTEST_HOP_PATH_ONLY
		cost[node2-'A'][node1-'A'] = 1;
		cost[node1-'A'][node2-'A'] = 1;
#endif
#ifdef MAXIMUM_FREE_CIRCUIT
		cost[node2-'A'][node1-'A'] = BIGGIE - available[node2-'A'][node1-'A'];
		cost[node1-'A'][node2-'A'] = BIGGIE - available[node1-'A'][node2-'A'];
#endif /* MFC */
#endif /* NEW */
	      } /* if available */
	  } /* if i > 0 */
      } /* for i = numhops */
  } /* end of ReleaseCall */

int main()
  {
    int row, col, i, j, callid, type;
    int delay, cap;
    char src, dst;
    float start, duration;
    
    activecalls = 0;
    fp1 = fopen("topology.dat", "r");
    if( fp1 == NULL )
      {
	fprintf(stderr, "Unable to open file topology.dat!\n");
	exit(1);
      }
    for( i = 0; i < MAX_ROW; i++ )
      for( j = 0; j < MAX_ROW; j++ )
	if( i == j )
	  cost[i][j] = 0;
	else cost[i][j] = INFINITY;
    for( i = 0; i < MAX_EVENTS; i++ )
      EventList[i].event_time = INFINITY;

    calls = 0;
    success = 0;
    blocked = 0;
    tothops = 0;
    totdelay = 0;
    numnodes = 0;
    numlinks = 0;
    while( (i = fscanf(fp1, "%c %c %d %d\n", &src, &dst, &delay, &cap)) == 4 )
      {
	float temp;
	row = src - 'A';
	col = dst - 'A';
	if( row > MAX_ROW )
	  {
	    printf("Too many rows to index!!\n");
	    exit(0);
	  }
	if( row >= numnodes )
	  numnodes = row + 1;
	if( col >= numnodes )
	  numnodes = col + 1;

	propdelay[row][col] = delay;
	propdelay[col][row] = delay;
	capacity[row][col] = cap;
	capacity[col][row] = cap;
	available[row][col] = cap;
	available[col][row] = cap;
#ifdef SHORTEST_DELAY_PATH_FIRST
	cost[row][col] = delay;
	cost[col][row] = delay;
#endif
#ifdef SHORTEST_HOP_PATH_FIRST
	cost[row][col] = 1;
	cost[col][row] = 1;
#endif
#ifdef SHORTEST_HOP_PATH_ONLY
	cost[row][col] = 1;
	cost[col][row] = 1;
#endif
#ifdef LEAST_LOADED_PATH
	temp = capacity[row][col];
	temp -= available[row][col];
	temp /= capacity[row][col];
	temp += DELTA;
#ifdef DEBUG3
	printf("Setting link cost from %c to %c to %4.3f\n",
	       row+'A', col+'A', temp);
#endif
	cost[row][col] = temp;
	cost[col][row] = temp;
#endif
#ifdef MAXIMUM_FREE_CIRCUIT
	cost[row][col] = BIGGIE - available[row][col];
	cost[col][row] = BIGGIE - available[col][row];
#endif
	numlinks++;
      }
    fclose(fp1);

#ifdef DEBUG
    printf("Read in a total of %d links\n", numlinks);
#endif

#ifdef SHORTEST_HOP_PATH_ONLY
    for( i = 0; i < numnodes; i++ )
      for( j = 0; j < numnodes; j++ )
	{
	  if( i == j )
	    mincost[i][j] = 0;
	  else
	    {
	      mincost[i][j] = DoDijkstra(i+'A', j+'A');
	      mincost[j][i] = mincost[i][j];
	    }
	}
#endif

    fp2 = fopen("workload.dat", "r");
    if( fp2 == NULL )
      {
	fprintf(stderr, "Unable to open file workload.dat!\n");
	exit(1);
      }

    /* Process the call workload file */
    numevents = 0;
    callid = 0;
    while( (i = fscanf(fp2, "%f %c %c %f\n", &start, &src, &dst, &duration)) == 4 )
      {
#ifdef DEBUG
	printf("Call %d from %c to %c starts at %8.6f lasts %8.6f\n",
	       callid, src, dst, start, duration);
#endif

	/* Enter this call arrival into the event list */
	EventList[numevents].event_type = CALL_ARRIVAL;
	EventList[numevents].event_time = start;
	EventList[numevents].source = src;
	EventList[numevents].destination = dst;
	EventList[numevents].callid = callid;
	EventList[numevents].duration = duration;
	numevents++;
	if( numevents == MAX_EVENTS )
	  {
	    printf("Ooops! Too many events to handle. Quitting.\n");
	    exit(0);
	  }

	callid++;
	calls++;
      }
    fclose(fp2);

    /* Now simulate the call arrivals and departures */
    current_event = 0;
    while( numevents > 0 )
      {
	numevents--;

	/* Get information about the current event */
	type = EventList[current_event].event_type;
	start = EventList[current_event].event_time;
	src = EventList[current_event].source;
	dst = EventList[current_event].destination;
	callid = EventList[current_event].callid;
	duration = EventList[current_event].duration;

#ifdef DEBUG
	printf("Event of type %d at time %8.6f (call %d from %c to %c)\n",
	       type, start, callid, src, dst);
#endif

	if( type == CALL_ARRIVAL )
	  {
	    now = start;
	    if( RouteCall(src, dst, callid) > 0 )
	      {
#ifdef DEBUG
		printf("Call succeeded!\n");
#endif
		activecalls++;
#ifdef DEBUG
		printf("%d\n", activecalls);
#endif
		success++;
		tothops += numhops - 1;
		totdelay += pathdelay;

		/* Enter this call completion into the event list */
		EventList[current_event].event_type = CALL_DEPARTURE;
		EventList[current_event].event_time = start + duration;
		EventList[current_event].callid = callid;
		EventList[current_event].source = src;
		EventList[current_event].destination = dst;
		EventList[current_event].duration = 0;
		numevents++;
		if( numevents == MAX_EVENTS )
		  {
		    printf("Ooops! Too many events to handle. Tired. Quitting.\n");
		    exit(0);
		  }

		/* sort the event list by time */
		qsort(EventList, MAX_EVENTS, sizeof(struct Event), mycompare);
	      }
	    else
	      {
#ifdef DEBUG
		printf("Call blocked!\n");
#endif
printf("Blocked call %d from %c to %c at time %4.3f\n", callid, src, dst, start);
		blocked++;
		current_event++;
	      }
	  }
	else if( type == CALL_DEPARTURE )
	  {
	    ReleaseCall(src, dst, callid);
	    activecalls--;
#ifdef DEBUG
	    printf("%d\n", activecalls);
#endif
	    current_event++;
	  }
	else
	  {
	    printf("Unknown event type %d\n", type);
	    current_event++;
	  }
      }
#ifdef SHORTEST_HOP_PATH_FIRST
    printf("Results for SHORTEST_HOP_PATH_FIRST\n");
#endif
#ifdef SHORTEST_HOP_PATH_ONLY
    printf("Results for SHORTEST_HOP_PATH_ONLY\n");
#endif
#ifdef SHORTEST_DELAY_PATH_FIRST
    printf("Results for SHORTEST_DELAY_PATH_FIRST\n");
#endif
#ifdef LEAST_LOADED_PATH
    printf("Results for LEAST_LOADED_PATH\n");
#endif
#ifdef MAXIMUM_FREE_CIRCUIT
    printf("Results for MAXIMUM_FREE_CIRCUIT\n");
#endif
    printf("Total calls: %d\n", calls);
    printf("Successful calls: %d\n", success);
    printf("Blocked calls: %d (%4.2f%%)\n",
	   blocked, (float) 100.0 * blocked/calls);
    printf("Mean hops per successful call: %5.3f\n",
	   (float) 1.0 * tothops/success);
    printf("Mean one-way delay per successful call: %5.3f\n",
	   (float) 1.0 * totdelay/success);

  }
